import unittest
import mock
from ...management.hooks import Hooks


class TestRules(unittest.TestCase):

    def test_init_with_optionals(self):
        t = Hooks(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')

        # with default params
        c.all()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/hooks', args[0])
        self.assertEqual(kwargs['params'], {
            "enabled": 'true',
            "fields": None,
            "include_fields": 'true',
            "page": None,
            "per_page": None,
            "include_totals": 'false',
        })

        # with fields params
        c.all(enabled=False, fields=['a', 'b'],
              include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/hooks', args[0])
        self.assertEqual(kwargs['params'], {'fields': 'a,b',
                                            'include_fields': 'false',
                                            'enabled': 'false',
                                            'page': None,
                                            'per_page': None,
                                            'include_totals': 'false'})

        # with pagination params
        c.all(page=3, per_page=27, include_totals=True)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/hooks', args[0])
        self.assertEqual(kwargs['params'], {'fields': None,
                                            'include_fields': 'true',
                                            'enabled': 'true',
                                            'page': 3,
                                            'per_page': 27,
                                            'include_totals': 'true'})

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.create({'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/hooks', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.get('an-id')

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id', args[0])
        self.assertEqual(kwargs['params'], {'fields': None})

        c.get('an-id', fields=['a', 'b'])

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id', args[0])
        self.assertEqual(kwargs['params'], {'fields': 'a,b'})

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.delete('an-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/hooks/an-id'
        )

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.update('an-id', {'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    # test for hooks secrets
    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_add_secret(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.add_secrets('an-id', {'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id/secrets', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_get_secrets(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.get_secrets('an-id')

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id/secrets', args[0])
        self.assertNotIn("params", kwargs)

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_delete_secrets(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.delete_secrets('an-id', ['a', 'b'])

        args, kwargs = mock_instance.delete.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id/secrets', args[0])
        self.assertEqual(kwargs['data'], ['a', 'b'])

    @mock.patch('auth0.v3.management.hooks.RestClient')
    def test_update_secrets(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Hooks(domain='domain', token='jwttoken')
        c.update_secrets('an-id', {'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual('https://domain/api/v2/hooks/an-id/secrets', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})



from .authorize_client import AuthorizeClient
from .database import Database
from .delegated import Delegated
from .enterprise import Enterprise
from .get_token import GetToken
from .logout import Logout
from .passwordless import Passwordless
from .social import Social
from .users import Users



from .base import AuthenticationBase


class Passwordless(AuthenticationBase):

    """Passwordless connections endpoints.

    Args:
        domain (str): Your auth0 domain (e.g: username.auth0.com)
    """

    def email(self, client_id, email, send='link', auth_params=None, client_secret=None):
        """Start flow sending an email.

        Given the user email address, it will send an email with:

          - A link (default, send:"link"). You can then authenticate with
            this user opening the link and he will be automatically logged in
            to the application. Optionally, you can append/override
            parameters to the link (like scope, redirect_uri, protocol,
            response_type, etc.) using auth_params dict.

          - A verification code (send:"code"). You can then authenticate with
            this user using email as username and code as password.

        Args:
            client_id (str): Client Id of the application.

            email (str): Email address.

            send (str, optional): Can be: 'link' or 'code'. Defaults to 'link'.

            auth_params (dict, optional): Parameters to append or override.

            client_secret (str): Client Secret of the application.
        """

        data={
            'client_id': client_id,
            'connection': 'email',
            'email': email,
            'send': send,
        }
        if auth_params:
            data.update({'authParams': auth_params})
        if client_secret:
            data.update({'client_secret': client_secret})

        return self.post(
            'https://{}/passwordless/start'.format(self.domain),
            data=data
        )

    def sms(self, client_id, phone_number, client_secret=None):
        """Start flow sending an SMS message.

        Given the user phone number, it will send an SMS with 
        a verification code. You can then authenticate with
        this user using phone number as username and code as password.

        Args:
            client_id (str): Client Id of the application.

            client_secret (str): Client Secret of the application.

            phone_number (str): Phone number.
        """

        data={
            'client_id': client_id,
            'connection': 'sms',
            'phone_number': phone_number,
        }
        if client_secret:
            data.update({'client_secret': client_secret})
            
        return self.post(
            'https://{}/passwordless/start'.format(self.domain),
            data=data
        )

    def sms_login(self, client_id, phone_number, code, scope='openid'):
        """Login using phone number/verification code.

        Args:
            client_id (str): Client Id of the application.

            phone_number (str): Phone number.

            code (str): Code received in the SMS.

            scope (str, optional): Scope to use. Defaults to 'openid'.
        """

        return self.post(
            'https://{}/oauth/ro'.format(self.domain),
            data={
                'client_id': client_id,
                'connection': 'sms',
                'grant_type': 'password',
                'username': phone_number,
                'password': code,
                'scope': scope,
            }
        )



import unittest
import mock
from ...management.custom_domains import CustomDomains


class TestCustomDomains(unittest.TestCase):

    def test_init_with_optionals(self):
        t = CustomDomains(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.custom_domains.RestClient')
    def test_get_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain='domain', token='jwttoken')
        g.all()

        mock_instance.get.assert_called_with(
            'https://domain/api/v2/custom-domains'
        )

    @mock.patch('auth0.v3.management.custom_domains.RestClient')
    def test_create_new(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain='domain', token='jwttoken')
        g.create_new(body={'a': 'b', 'c': 'd','e': 'f'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/custom-domains',args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd','e': 'f'})

    @mock.patch('auth0.v3.management.custom_domains.RestClient')
    def test_get_domain_by_id(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain='domain', token='jwttoken')
        g.get('an-id')

        mock_instance.get.assert_called_with('https://domain/api/v2/custom-domains/an-id')


    @mock.patch('auth0.v3.management.custom_domains.RestClient')
    def test_verify(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = CustomDomains(domain='domain', token='jwttoken')
        g.verify('an-id')

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/custom-domains/an-id/verify', args[0])





from .rest import RestClient


class EmailTemplates(object):
    """Auth0 email templates endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://{}/api/v2/email-templates'.format(self.domain)
        if id is not None:
            return '{}/{}'.format(url, id)
        return url

    def create(self, body):
        """Create a new email template.

        Args:
           body (dict): Attributes for the new email template.

        See: https://auth0.com/docs/api/management/v2#!/Email_Templates/post_email_templates
        """

        return self.client.post(self._url(), data=body)

    def get(self, template_name):
        """Retrieves an email template by its name.

        Args:
           template_name (str): Name of the email template to get.
              Must be one of: 'verify_email', 'reset_email', 'welcome_email',
              'blocked_account', 'stolen_credentials', 'enrollment_email',
              'change_password', 'password_reset', 'mfa_oob_code'.

        See: https://auth0.com/docs/api/management/v2#!/Email_Templates/get_email_templates_by_templateName
        """

        return self.client.get(self._url(template_name))

    def update(self, template_name, body):
        """Update an existing email template.

        Args:
           template_name (str): Name of the email template to update.
              Must be one of: 'verify_email', 'reset_email', 'welcome_email',
              'blocked_account', 'stolen_credentials', 'enrollment_email',
              'change_password', 'password_reset', 'mfa_oob_code'.

           body (dict): Attributes to update on the email template.

        See: https://auth0.com/docs/api/management/v2#!/Email_Templates/patch_email_templates_by_templateName
        """

        return self.client.patch(self._url(template_name), data=body)



import unittest
import mock
from ...management.emails import Emails


class TestEmails(unittest.TestCase):

    def test_init_with_optionals(self):
        t = Emails(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.emails.RestClient')
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        e = Emails(domain='domain', token='jwttoken')
        e.get()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/emails/provider', args[0])
        self.assertEqual(kwargs['params'], {'fields': None,
                                            'include_fields': 'true'})

        e.get(fields=['a', 'b'], include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/emails/provider', args[0])
        self.assertEqual(kwargs['params'], {'fields': 'a,b',
                                            'include_fields': 'false'})

    @mock.patch('auth0.v3.management.emails.RestClient')
    def test_config(self, mock_rc):
        mock_instance = mock_rc.return_value

        e = Emails(domain='domain', token='jwttoken')
        e.config({'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/emails/provider', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.emails.RestClient')
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        e = Emails(domain='domain', token='jwttoken')
        e.update({'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual('https://domain/api/v2/emails/provider', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.emails.RestClient')
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        e = Emails(domain='domain', token='jwttoken')
        e.delete()

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/emails/provider'
        )



from .rest import RestClient


class ResourceServers(object):
    """Auth0 resource servers endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://{}/api/v2/resource-servers'.format(self.domain)
        if id is not None:
            return '{}/{}'.format(url, id)
        return url

    def create(self, body):
        """Create a new resource server.

        Args:
           body (dict): Attributes for the new resource Server.

        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/post_resource_servers
        """

        return self.client.post(self._url(), data=body)

    def get_all(self, page=None, per_page=None, include_totals=False):
        """Retrieves all resource servers

        Args:
            page (int, optional): The result's page number (zero based). When not set,
              the default value is up to the server.

            per_page (int, optional): The amount of entries per page. When not set,
              the default value is up to the server.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to False.


        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/get_resource_servers
        """

        params = {
            'page': page,
            'per_page': per_page,
            'include_totals': str(include_totals).lower()
        }

        return self.client.get(self._url(), params=params)

    def get(self, id):
        """Retrieves a resource server by its id.

        Args:
           id (str): id of the resource server to get.


        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/get_resource_servers_by_id
        """

        return self.client.get(self._url(id))

    def delete(self, id):
        """Deletes a resource server.

        Args:
           id (str): Id of resource server to delete.


        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/delete_resource_servers_by_id
        """

        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Modifies a resource server.

        Args:
           id (str): The id of the resource server to update.

           body (dict): Attributes to modify.

        See: https://auth0.com/docs/api/management/v2#!/Resource_Servers/patch_resource_servers_by_id
        """

        return self.client.patch(self._url(id), data=body)



import unittest
import mock
from ...management.device_credentials import DeviceCredentials


class TestDeviceCredentials(unittest.TestCase):

    def test_init_with_optionals(self):
        t = DeviceCredentials(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.device_credentials.RestClient')
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = DeviceCredentials(domain='domain', token='jwttoken')
        c.get(user_id='uid', client_id='cid', type='type')

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/device-credentials', args[0])
        self.assertEqual(kwargs['params'], {'fields': None,
                                            'include_fields': 'true',
                                            'user_id': 'uid',
                                            'client_id': 'cid',
                                            'type': 'type'})

    @mock.patch('auth0.v3.management.device_credentials.RestClient')
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = DeviceCredentials(domain='domain', token='jwttoken')
        c.create({'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/device-credentials', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.device_credentials.RestClient')
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = DeviceCredentials(domain='domain', token='jwttoken')
        c.delete('an-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/device-credentials/an-id',
        )



import unittest
import mock
from ...management.users import Users


class TestUsers(unittest.TestCase):

    def test_init_with_optionals(self):
        t = Users(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_list(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.list()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/users', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 25,
            'page': 0,
            'include_totals': 'true',
            'sort': None,
            'connection': None,
            'fields': None,
            'include_fields': 'true',
            'q': None,
            'search_engine': None
        })

        u.list(page=1, per_page=50, sort='s', connection='con', q='q',
               search_engine='se', include_totals=False, fields=['a', 'b'],
               include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/users', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 50,
            'page': 1,
            'include_totals': 'false',
            'sort': 's',
            'connection': 'con',
            'fields': 'a,b',
            'include_fields': 'false',
            'q': 'q',
            'search_engine': 'se'
        })

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.create({'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/users', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_delete_all_users(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.delete_all_users()

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/users'
        )

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.get('an-id')

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/users/an-id', args[0])
        self.assertEqual(kwargs['params'], {'fields': None,
                                            'include_fields': 'true'})

        u.get('an-id', fields=['a', 'b'], include_fields=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/users/an-id', args[0])
        self.assertEqual(kwargs['params'], {'fields': 'a,b',
                                            'include_fields': 'false'})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.delete('an-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/users/an-id'
        )

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.update('an-id', {'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual('https://domain/api/v2/users/an-id', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_list_roles(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.list_roles('an-id')

        args, kwargs = mock_instance.get.call_args
        self.assertEqual('https://domain/api/v2/users/an-id/roles', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 25,
            'page': 0,
            'include_totals': 'true'
        })

        u.list_roles(id='an-id', page=1, per_page=50, include_totals=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/users/an-id/roles', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 50,
            'page': 1,
            'include_totals': 'false'
        })

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_remove_roles(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.remove_roles('an-id', ['a', 'b'])

        args, kwargs = mock_instance.delete.call_args

        self.assertEqual('https://domain/api/v2/users/an-id/roles',
                         args[0])
        self.assertEqual(kwargs['data'], {'roles': ['a', 'b']})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_add_roles(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.add_roles('an-id', ['a', 'b'])

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/users/an-id/roles',
                         args[0])
        self.assertEqual(kwargs['data'], {'roles': ['a', 'b']})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_list_permissions(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.list_permissions('an-id')

        args, kwargs = mock_instance.get.call_args
        self.assertEqual('https://domain/api/v2/users/an-id/permissions', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 25,
            'page': 0,
            'include_totals': 'true'
        })

        u.list_permissions(id='an-id', page=1, per_page=50, include_totals=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/users/an-id/permissions', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 50,
            'page': 1,
            'include_totals': 'false'
        })

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_remove_permissions(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.remove_permissions('an-id', ['a', 'b'])

        args, kwargs = mock_instance.delete.call_args

        self.assertEqual('https://domain/api/v2/users/an-id/permissions',
                         args[0])
        self.assertEqual(kwargs['data'], {'permissions': ['a', 'b']})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_add_permissions(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.add_permissions('an-id', ['a', 'b'])

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/users/an-id/permissions',
                         args[0])
        self.assertEqual(kwargs['data'], {'permissions': ['a', 'b']})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_delete_multifactor(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.delete_multifactor('an-id', 'provider')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/users/an-id/multifactor/provider'
        )

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_unlink_user_account(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.unlink_user_account('an-id', 'provider', 'user-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/users/an-id/identities/provider/user-id'
        )

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_link_user_account(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.link_user_account('user-id', {'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/users/user-id/identities',
                         args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_regenerate_recovery_code(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.regenerate_recovery_code('user-id')

        mock_instance.post.assert_called_with(
            'https://domain/api/v2/users/user-id/recovery-code-regeneration'
        )

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_get_guardian_enrollments(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.get_guardian_enrollments('user-id')

        mock_instance.get.assert_called_with(
            'https://domain/api/v2/users/user-id/enrollments'
        )

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_get_log_events(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.get_log_events('used_id')

        args, kwargs = mock_instance.get.call_args
        self.assertEqual('https://domain/api/v2/users/used_id/logs', args[0])
        self.assertEqual(kwargs['params']['page'], 0)
        self.assertEqual(kwargs['params']['per_page'], 50)
        self.assertIsNone(kwargs['params']['sort'])
        self.assertEqual(kwargs['params']['include_totals'], 'false')

    @mock.patch('auth0.v3.management.users.RestClient')
    def test_invalidate_remembered_browsers(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Users(domain='domain', token='jwttoken')
        u.invalidate_remembered_browsers('user-id')

        args, kwargs = mock_instance.post.call_args
        self.assertEqual('https://domain/api/v2/users/user-id/multifactor/actions/invalidate-remember-browser', args[0])



import base64
import json
from time import sleep

import mock
import sys
import requests
import unittest
from ...authentication.base import AuthenticationBase
from ...exceptions import Auth0Error, RateLimitError


class TestBase(unittest.TestCase):

    def test_telemetry_enabled_by_default(self):
        ab = AuthenticationBase('auth0.com')

        user_agent = ab.base_headers['User-Agent']
        auth0_client_bytes = base64.b64decode(ab.base_headers['Auth0-Client'])
        auth0_client_json = auth0_client_bytes.decode('utf-8')
        auth0_client = json.loads(auth0_client_json)
        content_type = ab.base_headers['Content-Type']

        from auth0 import __version__ as auth0_version
        python_version = '{}.{}.{}'.format(sys.version_info.major,
                                           sys.version_info.minor,
                                           sys.version_info.micro)

        client_info = {
            'name': 'auth0-python',
            'version': auth0_version,
            'env': {
                'python': python_version
            }
        }

        self.assertEqual(user_agent, 'Python/{}'.format(python_version))
        self.assertEqual(auth0_client, client_info)
        self.assertEqual(content_type, 'application/json')

    def test_telemetry_disabled(self):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        self.assertEqual(ab.base_headers, {'Content-Type': 'application/json'})

    @mock.patch('requests.post')
    def test_post(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False, timeout=(10, 2))

        mock_post.return_value.status_code = 200
        mock_post.return_value.text = '{"x": "y"}'

        data = ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

        mock_post.assert_called_with(url='the-url', json={'a': 'b'},
                                     headers={'c': 'd', 'Content-Type': 'application/json'}, timeout=(10, 2))

        self.assertEqual(data, {'x': 'y'})

    @mock.patch('requests.post')
    def test_post_with_defaults(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        mock_post.return_value.status_code = 200
        mock_post.return_value.text = '{"x": "y"}'

        # Only required params are passed
        data = ab.post('the-url')

        mock_post.assert_called_with(url='the-url', json=None,
                                     headers={'Content-Type': 'application/json'}, timeout=5.0)

        self.assertEqual(data, {'x': 'y'})

    @mock.patch('requests.post')
    def test_post_includes_telemetry(self, mock_post):
        ab = AuthenticationBase('auth0.com')

        mock_post.return_value.status_code = 200
        mock_post.return_value.text = '{"x": "y"}'

        data = ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

        self.assertEqual(mock_post.call_count, 1)
        call_kwargs = mock_post.call_args[1]
        self.assertEqual(call_kwargs['url'], 'the-url')
        self.assertEqual(call_kwargs['json'], {'a': 'b'})
        headers = call_kwargs['headers']
        self.assertEqual(headers['c'], 'd')
        self.assertEqual(headers['Content-Type'], 'application/json')
        self.assertIn('User-Agent', headers)
        self.assertIn('Auth0-Client', headers)

        self.assertEqual(data, {'x': 'y'})

    @mock.patch('requests.post')
    def test_post_error(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = '{"error": "e0",' \
                                          '"error_description": "desc"}'

            with self.assertRaises(Auth0Error) as context:
                ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code, 'e0')
            self.assertEqual(context.exception.message, 'desc')

    @mock.patch('requests.post')
    def test_post_rate_limit_error(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        mock_post.return_value.text = '{"statusCode": 429,' \
                                      ' "error": "e0",' \
                                      ' "error_description": "desc"}'
        mock_post.return_value.status_code = 429
        mock_post.return_value.headers = {
            'x-ratelimit-limit': '3',
            'x-ratelimit-remaining': '6',
            'x-ratelimit-reset': '9',
        }

        with self.assertRaises(Auth0Error) as context:
            ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, 'e0')
        self.assertEqual(context.exception.message, 'desc')
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, 9)

    @mock.patch('requests.post')
    def test_post_rate_limit_error_without_headers(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        mock_post.return_value.text = '{"statusCode": 429,' \
                                      ' "error": "e0",' \
                                      ' "error_description": "desc"}'
        mock_post.return_value.status_code = 429
        mock_post.return_value.headers = {}

        with self.assertRaises(Auth0Error) as context:
            ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

        self.assertEqual(context.exception.status_code, 429)
        self.assertEqual(context.exception.error_code, 'e0')
        self.assertEqual(context.exception.message, 'desc')
        self.assertIsInstance(context.exception, RateLimitError)
        self.assertEqual(context.exception.reset_at, -1)

    @mock.patch('requests.post')
    def test_post_error_with_code_property(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = '{"code": "e0",' \
                                          '"error_description": "desc"}'

            with self.assertRaises(Auth0Error) as context:
                ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code, 'e0')
            self.assertEqual(context.exception.message, 'desc')

    @mock.patch('requests.post')
    def test_post_error_with_no_error_code(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = '{"error_description": "desc"}'

            with self.assertRaises(Auth0Error) as context:
                ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code,
                             'a0.sdk.internal.unknown')
            self.assertEqual(context.exception.message, 'desc')

    @mock.patch('requests.post')
    def test_post_error_with_text_response(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = 'there has been a terrible error'

            with self.assertRaises(Auth0Error) as context:
                ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code,
                             'a0.sdk.internal.unknown')
            self.assertEqual(context.exception.message,
                             'there has been a terrible error')

    @mock.patch('requests.post')
    def test_post_error_with_no_response_text(self, mock_post):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        for error_status in [400, 500, None]:
            mock_post.return_value.status_code = error_status
            mock_post.return_value.text = None

            with self.assertRaises(Auth0Error) as context:
                ab.post('the-url', data={'a': 'b'}, headers={'c': 'd'})

            self.assertEqual(context.exception.status_code, error_status)
            self.assertEqual(context.exception.error_code,
                             'a0.sdk.internal.unknown')
            self.assertEqual(context.exception.message, '')

    @mock.patch('requests.get')
    def test_get(self, mock_get):
        ab = AuthenticationBase('auth0.com', telemetry=False, timeout=(10, 2))

        mock_get.return_value.status_code = 200
        mock_get.return_value.text = '{"x": "y"}'

        data = ab.get('the-url', params={'a': 'b'}, headers={'c': 'd'})

        mock_get.assert_called_with(url='the-url', params={'a': 'b'},
                                    headers={'c': 'd', 'Content-Type': 'application/json'}, timeout=(10, 2))

        self.assertEqual(data, {'x': 'y'})

    @mock.patch('requests.get')
    def test_get_with_defaults(self, mock_get):
        ab = AuthenticationBase('auth0.com', telemetry=False)

        mock_get.return_value.status_code = 200
        mock_get.return_value.text = '{"x": "y"}'

        # Only required params are passed
        data = ab.get('the-url')

        mock_get.assert_called_with(url='the-url', params=None,
                                    headers={'Content-Type': 'application/json'}, timeout=5.0)

        self.assertEqual(data, {'x': 'y'})

    @mock.patch('requests.get')
    def test_get_includes_telemetry(self, mock_get):
        ab = AuthenticationBase('auth0.com')

        mock_get.return_value.status_code = 200
        mock_get.return_value.text = '{"x": "y"}'

        data = ab.get('the-url', params={'a': 'b'}, headers={'c': 'd'})

        self.assertEqual(mock_get.call_count, 1)
        call_kwargs = mock_get.call_args[1]
        self.assertEqual(call_kwargs['url'], 'the-url')
        self.assertEqual(call_kwargs['params'], {'a': 'b'})
        headers = call_kwargs['headers']
        self.assertEqual(headers['c'], 'd')
        self.assertEqual(headers['Content-Type'], 'application/json')
        self.assertIn('User-Agent', headers)
        self.assertIn('Auth0-Client', headers)

        self.assertEqual(data, {"x": "y"})

    def test_get_can_timeout(self):
        ab = AuthenticationBase('auth0.com', timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            ab.get('https://google.com', params={'a': 'b'}, headers={'c': 'd'})

    def test_post_can_timeout(self):
        ab = AuthenticationBase('auth0.com', timeout=0.00001)

        with self.assertRaises(requests.exceptions.Timeout):
            ab.post('https://google.com', data={'a': 'b'}, headers={'c': 'd'})



import unittest
import mock
from ...management.grants import Grants


class TestGrants(unittest.TestCase):

    def test_init_with_optionals(self):
        t = Grants(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.grants.RestClient')
    def test_get_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = Grants(domain='domain', token='jwttoken')
        g.all(extra_params={'user_id':'an-id', 'client_id': 'an-id', 'audience':'test'})

        args, kwargs = mock_instance.get.call_args

        mock_instance.get.assert_called_with(
            'https://domain/api/v2/grants', params={'user_id': 'an-id', 'client_id': 'an-id', 'audience': 'test', 'page': None, 'per_page': None, 'include_totals': 'false'}
        )


    @mock.patch('auth0.v3.management.grants.RestClient')
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = Grants(domain='domain', token='jwttoken')
        c.delete('an-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/grants/an-id'
        )



import base64
import json
import sys
import platform
import requests
from ..exceptions import Auth0Error, RateLimitError

UNKNOWN_ERROR = 'a0.sdk.internal.unknown'


class AuthenticationBase(object):
    """Base authentication object providing simple REST methods.

    Args:
        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)
        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, telemetry=True, timeout=5.0):
        self.domain = domain
        self.timeout = timeout
        self.base_headers = {'Content-Type': 'application/json'}

        if telemetry:
            py_version = platform.python_version()
            version = sys.modules['auth0'].__version__

            auth0_client = json.dumps({
                'name': 'auth0-python',
                'version': version,
                'env': {
                    'python': py_version,
                }
            }).encode('utf-8')

            self.base_headers.update({
                'User-Agent': 'Python/{}'.format(py_version),
                'Auth0-Client': base64.b64encode(auth0_client),
            })

    def post(self, url, data=None, headers=None):
        request_headers = self.base_headers.copy()
        request_headers.update(headers or {})
        response = requests.post(url=url, json=data, headers=request_headers, timeout=self.timeout)
        return self._process_response(response)

    def get(self, url, params=None, headers=None):
        request_headers = self.base_headers.copy()
        request_headers.update(headers or {})
        response = requests.get(url=url, params=params, headers=request_headers, timeout=self.timeout)
        return self._process_response(response)

    def _process_response(self, response):
        return self._parse(response).content()

    def _parse(self, response):
        if not response.text:
            return EmptyResponse(response.status_code)
        try:
            return JsonResponse(response)
        except ValueError:
            return PlainResponse(response)


class Response(object):
    def __init__(self, status_code, content, headers):
        self._status_code = status_code
        self._content = content
        self._headers = headers

    def content(self):
        if self._is_error():
            if self._status_code == 429:
                reset_at = int(self._headers.get('x-ratelimit-reset', '-1'))
                raise RateLimitError(error_code=self._error_code(),
                                     message=self._error_message(),
                                     reset_at=reset_at)

            raise Auth0Error(status_code=self._status_code,
                             error_code=self._error_code(),
                             message=self._error_message())
        else:
            return self._content

    def _is_error(self):
        return self._status_code is None or self._status_code >= 400

    # Adding these methods to force implementation in subclasses because they are references in this parent class
    def _error_code(self):
        raise NotImplementedError

    def _error_message(self):
        raise NotImplementedError


class JsonResponse(Response):
    def __init__(self, response):
        content = json.loads(response.text)
        super(JsonResponse, self).__init__(response.status_code, content, response.headers)

    def _error_code(self):
        if 'error' in self._content:
            return self._content.get('error')
        elif 'code' in self._content:
            return self._content.get('code')
        else:
            return UNKNOWN_ERROR

    def _error_message(self):
        return self._content.get('error_description', '')


class PlainResponse(Response):
    def __init__(self, response):
        super(PlainResponse, self).__init__(response.status_code, response.text, response.headers)

    def _error_code(self):
        return UNKNOWN_ERROR

    def _error_message(self):
        return self._content


class EmptyResponse(Response):
    def __init__(self, status_code):
        super(EmptyResponse, self).__init__(status_code, '', {})

    def _error_code(self):
        return UNKNOWN_ERROR

    def _error_message(self):
        return ''



from .base import AuthenticationBase
import warnings


class Users(AuthenticationBase):

    """Userinfo related endpoints.

    Args:
        domain (str): Your auth0 domain (e.g: username.auth0.com)
    """

    def userinfo(self, access_token):

        """Returns the user information based on the Auth0 access token.
        This endpoint will work only if openid was granted as a scope for the access_token.

        Args:
            access_token (str): Auth0 access token (obtained during login).

        Returns:
            The user profile.
        """

        return self.get(
            url='https://{}/userinfo'.format(self.domain),
            headers={'Authorization': 'Bearer {}'.format(access_token)}
        )

    def tokeninfo(self, jwt):

        """Returns user profile based on the user's jwt

        Validates a JSON Web Token (signature and expiration) and returns the
        user information associated with the user id (sub property) of
        the token.

        Args:
            jwt (str): User's jwt

        Returns:
            The user profile.
        """
        warnings.warn("/tokeninfo will be deprecated in future releases", DeprecationWarning)
        return self.post(
            url='https://{}/tokeninfo'.format(self.domain),
            data={'id_token': jwt}
        )



from .base import AuthenticationBase


class Delegated(AuthenticationBase):
    """Delegated authentication endpoints.

    Args:
        domain (str): Your auth0 domain (e.g: username.auth0.com)
    """

    def get_token(self, client_id, target, api_type, grant_type,
                  id_token=None, refresh_token=None, scope='openid'):

        """Obtain a delegation token.
        """

        if id_token and refresh_token:
            raise ValueError('Only one of id_token or refresh_token '
                             'can be None')

        data = {
            'client_id': client_id,
            'grant_type': grant_type,
            'target': target,
            'scope': scope,
            'api_type': api_type,
        }

        if id_token:
            data.update({'id_token': id_token})
        elif refresh_token:
            data.update({'refresh_token': refresh_token})
        else:
            raise ValueError('Either id_token or refresh_token must '
                             'have a value')

        return self.post('https://{}/delegation'.format(self.domain), data=data)



from .rest import RestClient


class Blacklists(object):
    """Auth0 blacklists endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.url = 'https://{}/api/v2/blacklists/tokens'.format(domain)
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def get(self, aud=None):
        """Retrieves the jti and aud of all tokens in the blacklist.

        Args:
            aud (str, optional): The JWT's aud claim. The client_id of the
                application for which it was issued.


        See: https://auth0.com/docs/api/management/v2#!/Blacklists/get_tokens
        """

        params = {
            'aud': aud
        }

        return self.client.get(self.url, params=params)

    def create(self, jti, aud=None):
        """Adds a token to the blacklist.

        Args:
            jti (str): the jti of the JWT to blacklist.

            aud (str, optional): The JWT's aud claim. The client_id of the
                application for which it was issued.

        See: https://auth0.com/docs/api/management/v2#!/Blacklists/post_tokens
        """
        body = {
            'jti': jti,
        }

        if aud:
            body.update({'aud': aud})

        return self.client.post(self.url, data=body)



from .rest import RestClient


class LogStreams(object):
    """Auth0 log streams endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://{}/api/v2/log-streams'.format(self.domain)
        if id is not None:
            return '{}/{}'.format(url, id)
        return url

    def list(self):
        """Search log events.

        Args:
        See: https://auth0.com/docs/api/management/v2/#!/Log_Streams/get_log_streams
        """

        return self.client.get(self._url())

    def get(self, id):
        """Retrieves the data related to the log stream entry identified by id.

        Args:
            id (str): The id of the log stream to retrieve.

        See: https://auth0.com/docs/api/management/v2/#!/Log_Streams/get_log_streams_by_id
        """

        return self.client.get(self._url(id))

    def create(self, body):
        """Creates a new log stream.

        Args:
            body (dict): the attributes for the role to create.

        See: https://auth0.com/docs/api/management/v2/#!/Log_Streams/post_log_streams
        """
        return self.client.post(self._url(), data=body)

    def delete(self, id):
        """Delete a log stream.

        Args:
            id (str): The id of the log ste to delete.

        See: https://auth0.com/docs/api/management/v2/#!/Log_Streams/delete_log_streams_by_id
        """
        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Update a log stream with the attributes passed in 'body'

        Args:
            id (str): The id of the log stream to update.

            body (dict): the attributes to update on the log stream.

        See: https://auth0.com/docs/api/management/v2/#!/Log_Streams/patch_log_streams_by_id
        """
        return self.client.patch(self._url(id), data=body)



from .rest import RestClient


class Stats(object):
    """Auth0 stats endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, action):
        return 'https://{}/api/v2/stats/{}'.format(self.domain, action)

    def active_users(self):
        """Gets the active users count (logged in during the last 30 days).

        Returns: An integer.

        See: https://auth0.com/docs/api/management/v2#!/Stats/get_active_users
        """

        return self.client.get(self._url('active-users'))

    def daily_stats(self, from_date=None, to_date=None):
        """Gets the daily stats for a particular period.

        Args:
           from_date (str, optional): The first day of the period (inclusive) in
              YYYYMMDD format.

           to_date (str, optional): The last day of the period (inclusive) in
              YYYYMMDD format.

        See: https://auth0.com/docs/api/management/v2#!/Stats/get_daily
        """

        return self.client.get(self._url('daily'), params={'from': from_date,
                                                           'to': to_date})



import unittest
import mock
from ...management.rules_configs import RulesConfigs


class TestRulesConfigs(unittest.TestCase):

    def test_init_with_optionals(self):
        t = RulesConfigs(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.rules_configs.RestClient')
    def test_all(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = RulesConfigs(domain='domain', token='jwttoken')

        c.all()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/rules-configs', args[0])

    @mock.patch('auth0.v3.management.rules_configs.RestClient')
    def test_unset(self, mock_rc):
        mock_instance = mock_rc.return_value

        c = RulesConfigs(domain='domain', token='jwttoken')
        c.unset('an-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/rules-configs/an-id'
        )

    @mock.patch('auth0.v3.management.rules_configs.RestClient')
    def test_set(self, mock_rc):
        mock_instance = mock_rc.return_value

        g = RulesConfigs(domain='domain', token='jwttoken')
        g.set('key', 'MY_RULES_CONFIG_VALUES')

        args, kwargs = mock_instance.put.call_args
        self.assertEqual('https://domain/api/v2/rules-configs/key', args[0])







from .rest import RestClient


class CustomDomains(object):
    """Auth0 custom domains endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://%s/api/v2/custom-domains' % self.domain
        if id is not None:
            return url + '/' + id
        return url

    def all(self):
        """Retrieves all custom domains.

        See: https://auth0.com/docs/api/management/v2#!/Custom_Domains/get_custom_domains
        """
        return self.client.get(self._url())

    def get(self, id):
        """Retrieves custom domain.

        See: https://auth0.com/docs/api/management/v2#!/Custom_Domains/get_custom_domains_by_id
        """
        url = self._url('%s' % (id))
        return self.client.get(url)

    def delete(self, id):
        """Deletes a grant.

        Args:
           id (str): The id of the custom domain to delete.

        See: https://auth0.com/docs/api/management/v2#!/Custom_Domains/delete_custom_domains_by_id
        """
        url = self._url('%s' % (id))
        return self.client.delete(url)

    def create_new(self, body):
        """Configure a new custom domain.

        Args:
           body (str): The domain, tye and verification method in json.

        See: https://auth0.com/docs/api/management/v2#!/Custom_Domains/post_custom_domains
        """
        return self.client.post(self._url(), data=body)

    def verify(self, id):
        """Verify a custom domain.

        Args:
           id (str): The id of the custom domain to delete.

        See: https://auth0.com/docs/api/management/v2#!/Custom_Domains/post_verify
        """
        url = self._url('%s/verify' % (id))
        return self.client.post(url)



class Auth0Error(Exception):
    def __init__(self, status_code, error_code, message):
        self.status_code = status_code
        self.error_code = error_code
        self.message = message

    def __str__(self):
        return '{}: {}'.format(self.status_code, self.message)


class RateLimitError(Auth0Error):
    def __init__(self, error_code, message, reset_at):
        super(RateLimitError, self).__init__(status_code=429, error_code=error_code, message=message)
        self.reset_at = reset_at


class TokenValidationError(Exception):
    pass



import unittest
import mock
from ...management.roles import Roles


class TestRoles(unittest.TestCase):

    def test_init_with_optionals(self):
        t = Roles(domain='domain', token='jwttoken', telemetry=False, timeout=(10, 2))
        self.assertEqual(t.client.timeout, (10, 2))
        telemetry_header = t.client.base_headers.get('Auth0-Client', None)
        self.assertEqual(telemetry_header, None)

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_list(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.list()

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/roles', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 25,
            'page': 0,
            'include_totals': 'true',
            'name_filter': None
        })

        u.list(page=1, per_page=50, include_totals=False, name_filter='little-role')

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/roles', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 50,
            'page': 1,
            'include_totals': 'false',
            'name_filter': 'little-role'
        })

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_create(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.create({'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/roles', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_get(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.get('an-id')

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id', args[0])

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_delete(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.delete('an-id')

        mock_instance.delete.assert_called_with(
            'https://domain/api/v2/roles/an-id'
        )

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_update(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.update('an-id', {'a': 'b', 'c': 'd'})

        args, kwargs = mock_instance.patch.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id', args[0])
        self.assertEqual(kwargs['data'], {'a': 'b', 'c': 'd'})

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_list_users(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.list_users('an-id')

        args, kwargs = mock_instance.get.call_args
        self.assertEqual('https://domain/api/v2/roles/an-id/users', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 25,
            'page': 0,
            'include_totals': 'true'
        })

        u.list_users(id='an-id', page=1, per_page=50, include_totals=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id/users', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 50,
            'page': 1,
            'include_totals': 'false'
        })


    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_add_users(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.add_users('an-id', ['a', 'b'])

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id/users',
                         args[0])
        self.assertEqual(kwargs['data'], {'users': ['a', 'b']})

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_list_permissions(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.list_permissions('an-id')

        args, kwargs = mock_instance.get.call_args
        self.assertEqual('https://domain/api/v2/roles/an-id/permissions', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 25,
            'page': 0,
            'include_totals': 'true'
        })

        u.list_permissions(id='an-id', page=1, per_page=50, include_totals=False)

        args, kwargs = mock_instance.get.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id/permissions', args[0])
        self.assertEqual(kwargs['params'], {
            'per_page': 50,
            'page': 1,
            'include_totals': 'false'
        })

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_remove_permissions(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.remove_permissions('an-id', ['a', 'b'])

        args, kwargs = mock_instance.delete.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id/permissions',
                         args[0])
        self.assertEqual(kwargs['data'], {'permissions': ['a', 'b']})

    @mock.patch('auth0.v3.management.roles.RestClient')
    def test_add_permissions(self, mock_rc):
        mock_instance = mock_rc.return_value

        u = Roles(domain='domain', token='jwttoken')
        u.add_permissions('an-id', ['a', 'b'])

        args, kwargs = mock_instance.post.call_args

        self.assertEqual('https://domain/api/v2/roles/an-id/permissions',
                         args[0])
        self.assertEqual(kwargs['data'], {'permissions': ['a', 'b']})



import unittest
import mock
from ...authentication.users import Users


class TestUsers(unittest.TestCase):

    @mock.patch('auth0.v3.authentication.users.Users.get')
    def test_userinfo(self, mock_get):

        u = Users('my.domain.com')

        u.userinfo(access_token='atk')

        mock_get.assert_called_with(
            url='https://my.domain.com/userinfo',
            headers={'Authorization': 'Bearer atk'}
        )

    @mock.patch('auth0.v3.authentication.users.Users.post')
    def test_tokeninfo(self, mock_post):

        u = Users('my.domain.com')

        u.tokeninfo(jwt='jwtoken')

        mock_post.assert_called_with(
            url='https://my.domain.com/tokeninfo',
            data={'id_token': 'jwtoken'}
        )



from .rest import RestClient


class Tickets(object):
    """Auth0 tickets endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, action):
        return 'https://{}/api/v2/tickets/{}'.format(self.domain, action)

    def create_email_verification(self, body):
        """Create an email verification ticket.

        Args:
            body (dict): attributes to set on the email verification request.

        See: https://auth0.com/docs/api/v2#!/Tickets/post_email_verification
        """
        return self.client.post(self._url('email-verification'), data=body)

    def create_pswd_change(self, body):
        """Create password change ticket.

        Args:
            body (dict): attributes to set on the password change request.

        See: https://auth0.com/docs/api/v2#!/Tickets/post_password_change
        """
        return self.client.post(self._url('password-change'), data=body)



import json
import time

import jwt
import requests

from auth0.v3.exceptions import TokenValidationError


class SignatureVerifier(object):
    DISABLE_JWT_CHECKS = {
        "verify_signature": True,
        "verify_exp": False,
        "verify_nbf": False,
        "verify_iat": False,
        "verify_aud": False,
        "verify_iss": False,
        "require_exp": False,
        "require_iat": False,
        "require_nbf": False,
    }

    """Abstract class that will verify a given JSON web token's signature
    using the key fetched internally given its key id.
    
    Args:
        algorithm (str): The expected signing algorithm (e.g. RS256).
    """

    def __init__(self, algorithm):
        if not algorithm or type(algorithm) != str:
            raise ValueError("algorithm must be specified.")
        self._algorithm = algorithm

    """Obtains the key associated to the given key id.
    Must be implemented by subclasses.

    Args:
        key_id (str, optional): The id of the key to fetch.
    
    Returns:
        the key to use for verifying a cryptographic signature
    """

    def _fetch_key(self, key_id=None):
        raise NotImplementedError

    """Verifies the signature of the given JSON web token.

    Args:
        token (str): The JWT to get its signature verified.

    Raises:
        TokenValidationError: if the token cannot be decoded, the algorithm is invalid
        or the token's signature doesn't match the calculated one.
    """

    def verify_signature(self, token):
        try:
            header = jwt.get_unverified_header(token)
        except jwt.exceptions.DecodeError:
            raise TokenValidationError("ID token could not be decoded.")

        alg = header.get('alg', None)
        if alg != self._algorithm:
            raise TokenValidationError(
                'Signature algorithm of "{}" is not supported. Expected the ID token '
                'to be signed with "{}"'.format(alg, self._algorithm))

        kid = header.get('kid', None)
        secret_or_certificate = self._fetch_key(key_id=kid)

        try:
            decoded = jwt.decode(jwt=token, key=secret_or_certificate,
                                 algorithms=[self._algorithm], options=self.DISABLE_JWT_CHECKS)
        except jwt.exceptions.InvalidSignatureError:
            raise TokenValidationError("Invalid token signature.")
        return decoded


class SymmetricSignatureVerifier(SignatureVerifier):
    """Verifier for HMAC signatures, which rely on shared secrets.

    Args:
        shared_secret (str): The shared secret used to decode the token.
        algorithm (str, optional): The expected signing algorithm. Defaults to "HS256".
    """

    def __init__(self, shared_secret, algorithm="HS256"):
        super(SymmetricSignatureVerifier, self).__init__(algorithm)
        self._shared_secret = shared_secret

    def _fetch_key(self, key_id=None):
        return self._shared_secret


class AsymmetricSignatureVerifier(SignatureVerifier):
    """Verifier for RSA signatures, which rely on public key certificates.

    Args:
        jwks_url (str): The url where the JWK set is located.
        algorithm (str, optional): The expected signing algorithm. Defaults to "RS256".
    """

    def __init__(self, jwks_url, algorithm="RS256"):
        super(AsymmetricSignatureVerifier, self).__init__(algorithm)
        self._fetcher = JwksFetcher(jwks_url)

    def _fetch_key(self, key_id=None):
        return self._fetcher.get_key(key_id)


class JwksFetcher(object):
    CACHE_TTL = 600  # 10 min cache lifetime

    """Class that fetches and holds a JSON web key set.
    This class makes use of an in-memory cache. For it to work properly, define this instance once and re-use it.

    Args:
        jwks_url (str): The url where the JWK set is located.
        cache_ttl (str, optional): The lifetime of the JWK set cache in seconds. Defaults to 600 seconds.
    """

    def __init__(self, jwks_url, cache_ttl=CACHE_TTL):
        self._jwks_url = jwks_url
        self._init_cache(cache_ttl)
        return

    def _init_cache(self, cache_ttl):
        self._cache_value = {}
        self._cache_date = 0
        self._cache_ttl = cache_ttl
        self._cache_is_fresh = False

    """Attempts to obtain the JWK set from the cache, as long as it's still valid.
    When not, it will perform a network request to the jwks_url to obtain a fresh result
    and update the cache value with it.

    Args:
        force (bool, optional): whether to ignore the cache and force a network request or not. Defaults to False.
    """

    def _fetch_jwks(self, force=False):
        has_expired = self._cache_date + self._cache_ttl < time.time()

        if not force and not has_expired:
            # Return from cache
            self._cache_is_fresh = False
            return self._cache_value

        # Invalidate cache and fetch fresh data
        self._cache_value = {}
        response = requests.get(self._jwks_url)

        if response.ok:
            # Update cache
            jwks = response.json()
            self._cache_value = self._parse_jwks(jwks)
            self._cache_is_fresh = True
            self._cache_date = time.time()
        return self._cache_value

    """Converts a JWK string representation into a binary certificate in PEM format.
    """

    @staticmethod
    def _parse_jwks(jwks):
        keys = {}

        for key in jwks['keys']:
            # noinspection PyUnresolvedReferences
            # requirement already includes cryptography -> pyjwt[crypto]
            rsa_key = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(key))
            keys[key["kid"]] = rsa_key
        return keys

    """Obtains the JWK associated with the given key id.

    Args:
        key_id (str): The id of the key to fetch.

    Returns:
        the JWK associated with the given key id.
    
    Raises:
        TokenValidationError: when a key with that id cannot be found
    """

    def get_key(self, key_id):
        keys = self._fetch_jwks()

        if keys and key_id in keys:
            return keys[key_id]

        if not self._cache_is_fresh:
            keys = self._fetch_jwks(force=True)
            if keys and key_id in keys:
                return keys[key_id]
        raise TokenValidationError('RSA Public Key with ID "{}" was not found.'.format(key_id))


class TokenVerifier():
    """Class that verifies ID tokens following the steps defined in the OpenID Connect spec.
    An OpenID Connect ID token is not meant to be consumed until it's verified.

    Args:
        signature_verifier (SignatureVerifier): The instance that knows how to verify the signature.
        issuer (str): The expected issuer claim value.
        audience (str): The expected audience claim value.
        leeway (int, optional): The clock skew to accept when verifying date related claims in seconds.
            Defaults to 60 seconds.
    """

    def __init__(self, signature_verifier, issuer, audience, leeway=0):
        if not signature_verifier or not isinstance(signature_verifier, SignatureVerifier):
            raise TypeError("signature_verifier must be an instance of SignatureVerifier.")

        self.iss = issuer
        self.aud = audience
        self.leeway = leeway
        self._sv = signature_verifier
        self._clock = None  # visible for testing

    """Attempts to verify the given ID token, following the steps defined in the OpenID Connect spec.

    Args:
        token (str): The JWT to verify.
        nonce (str, optional): The nonce value sent during authentication.
        max_age (int, optional): The max_age value sent during authentication.
        organization (str, optional): The expected organization ID (org_id) claim value. This should be specified
            when logging in to an organization.
        
    Raises:
        TokenValidationError: when the token cannot be decoded, the token signing algorithm is not the expected one, 
        the token signature is invalid or the token has a claim missing or with unexpected value.
    """

    def verify(self, token, nonce=None, max_age=None, organization=None):
        # Verify token presence
        if not token or not isinstance(token, str):
            raise TokenValidationError("ID token is required but missing.")

        # Verify algorithm and signature
        payload = self._sv.verify_signature(token)

        # Verify claims
        self._verify_payload(payload, nonce, max_age, organization)

    def _verify_payload(self, payload, nonce=None, max_age=None, organization=None):
        try:
            # on Python 2.7, 'str' keys as parsed as 'unicode'
            # But 'unicode' was removed on Python 3.7
            # noinspection PyUnresolvedReferences
            ustr = unicode
        except NameError:
            ustr = str

        # Issuer
        if 'iss' not in payload or not isinstance(payload['iss'], (str, ustr)):
            raise TokenValidationError('Issuer (iss) claim must be a string present in the ID token')
        if payload['iss'] != self.iss:
            raise TokenValidationError(
                'Issuer (iss) claim mismatch in the ID token; expected "{}", '
                'found "{}"'.format(self.iss, payload['iss']))

        # Subject
        if 'sub' not in payload or not isinstance(payload['sub'], (str, ustr)):
            raise TokenValidationError('Subject (sub) claim must be a string present in the ID token')

        # Audience
        if 'aud' not in payload or not (isinstance(payload['aud'], (str, ustr)) or isinstance(payload['aud'], list)):
            raise TokenValidationError(
                'Audience (aud) claim must be a string or array of strings present in the ID token')

        if isinstance(payload['aud'], list) and not self.aud in payload['aud']:
            payload_audiences = ", ".join(payload['aud'])
            raise TokenValidationError(
                'Audience (aud) claim mismatch in the ID token; expected "{}" but was '
                'not one of "{}"'.format(self.aud, payload_audiences))
        elif isinstance(payload['aud'], (str, ustr)) and payload['aud'] != self.aud:
            raise TokenValidationError(
                'Audience (aud) claim mismatch in the ID token; expected "{}" '
                'but found "{}"'.format(self.aud, payload['aud']))

        # --Time validation (epoch)--
        now = self._clock or time.time()
        leeway = self.leeway

        # Expires at
        if 'exp' not in payload or not isinstance(payload['exp'], int):
            raise TokenValidationError('Expiration Time (exp) claim must be a number present in the ID token')

        exp_time = payload['exp'] + leeway
        if now > exp_time:
            raise TokenValidationError(
                'Expiration Time (exp) claim error in the ID token; current time ({}) is '
                'after expiration time ({})'.format(now, exp_time))

        # Issued at
        if 'iat' not in payload or not isinstance(payload['iat'], int):
            raise TokenValidationError('Issued At (iat) claim must be a number present in the ID token')

        # Nonce
        if nonce:
            if 'nonce' not in payload or not isinstance(payload['nonce'], (str, ustr)):
                raise TokenValidationError('Nonce (nonce) claim must be a string present in the ID token')
            if payload['nonce'] != nonce:
                raise TokenValidationError(
                    'Nonce (nonce) claim mismatch in the ID token; expected "{}", '
                    'found "{}"'.format(nonce, payload['nonce']))

        # Organization
        if organization:
            if 'org_id' not in payload or not isinstance(payload['org_id'], (str, ustr)):
                raise TokenValidationError('Organization (org_id) claim must be a string present in the ID token')
            if payload['org_id'] != organization:
                raise TokenValidationError(
                    'Organization (org_id) claim mismatch in the ID token; expected "{}", '
                    'found "{}"'.format(organization, payload['org_id']))

        # Authorized party
        if isinstance(payload['aud'], list) and len(payload['aud']) > 1:
            if 'azp' not in payload or not isinstance(payload['azp'], (str, ustr)):
                raise TokenValidationError(
                    'Authorized Party (azp) claim must be a string present in the ID token when '
                    'Audience (aud) claim has multiple values')
            if payload['azp'] != self.aud:
                raise TokenValidationError(
                    'Authorized Party (azp) claim mismatch in the ID token; expected "{}", '
                    'found "{}"'.format(self.aud, payload['azp']))

        # Authentication time
        if max_age:
            if 'auth_time' not in payload or not isinstance(payload['auth_time'], int):
                raise TokenValidationError(
                    'Authentication Time (auth_time) claim must be a number present in the ID token '
                    'when Max Age (max_age) is specified')

            auth_valid_until = payload['auth_time'] + max_age + leeway
            if now > auth_valid_until:
                raise TokenValidationError(
                    'Authentication Time (auth_time) claim in the ID token indicates that too much '
                    'time has passed since the last end-user authentication. Current time ({}) '
                    'is after last auth at ({})'.format(now, auth_valid_until))



from .rest import RestClient


class Hooks(object):

    """Hooks endpoint implementation.

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = "https://{}/api/v2/hooks".format(self.domain)
        if id is not None:
            return "{}/{}".format(url, id)
        return url

    def all(
        self,
        enabled=True,
        fields=None,
        include_fields=True,
        page=None,
        per_page=None,
        include_totals=False,
    ):
        """Retrieves a list of all hooks.

        Args:
            enabled (bool, optional): If provided, retrieves hooks that match
                the value, otherwise all hooks are retrieved.

            fields (list, optional): A list of fields to include or exclude
                (depending on include_fields) from the result, empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise
                (defaults to true).

            page (int, optional): The result's page number (zero based).

            per_page (int, optional): The amount of entries per page.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/get_hooks
        """

        params = {
            "fields": fields and ",".join(fields) or None,
            "include_fields": str(include_fields).lower(),
            "page": page,
            "per_page": per_page,
            "include_totals": str(include_totals).lower(),
        }

        # since the default is True, this is here to disable the filter
        if enabled is not None:
            params["enabled"] = str(enabled).lower()

        return self.client.get(self._url(), params=params)

    def create(self, body):
        """Creates a new Hook.

        Args:
            body (dict): Attributes for the newly created hook,
                See: https://auth0.com/docs/api/v2#!/Hooks/post_hooks
        """
        return self.client.post(self._url(), data=body)

    def get(self, id, fields=None):
        """Retrieves a hook by its ID.

        Args:
            id (str): The id of the hook to retrieve.

            fields (list, optional): A list of fields to include or exclude
                (depending on include_fields) from the result, empty to
                retrieve all fields.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/get_hooks_by_id
        """
        params = {
            "fields": fields and ",".join(fields) or None,
        }
        return self.client.get(self._url(id), params=params)

    def delete(self, id):
        """Deletes a hook.

        Args:
            id (str): The id of the hook to delete.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/delete_hooks_by_id
        """
        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Updates an existing hook.

        Args:
            id (str): The id of the hook to modify.

            body (dict): Attributes to modify.

        See: https://auth0.com/docs/api/v2#!/Hooks/patch_hooks_by_id
        """
        return self.client.patch(self._url(id), data=body)

    def get_secrets(self, id):
        """Retrieves a hook's secrets.

        Args:
            id (str): The id of the hook to retrieve secrets from.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/get_secrets
        """

        return self.client.get(self._url("%s/secrets" % id))

    def add_secrets(self, id, body):
        """Add one or more secrets for an existing hook.

        Args:
            id (str): The id of the hook to add secrets to.

            body (dict): Dict of key-value pairs where the value must be a string.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/post_secrets
        """
        return self.client.post(self._url("%s/secrets" % id), data=body)

    def delete_secrets(self, id, body):
        """Delete one or more existing secrets for an existing hook.

        Args:
            id (str): The id of the hook to add secrets to.

            body (list): List of secret names to delete.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/delete_secrets
        """
        return self.client.delete(self._url("%s/secrets" % id), data=body)

    def update_secrets(self, id, body):
        """Update one or more existing secrets for an existing hook.

        Args:
            id (str): The id of the hook to add secrets to.

            body (dict): Dict of key-value pairs where the value must be a string.

        See: https://auth0.com/docs/api/management/v2#!/Hooks/patch_secrets
        """
        return self.client.patch(self._url("%s/secrets" % id), data=body)



from .blacklists import Blacklists
from .client_grants import ClientGrants
from .clients import Clients
from .connections import Connections
from .custom_domains import CustomDomains
from .device_credentials import DeviceCredentials
from .email_templates import EmailTemplates
from .emails import Emails
from .grants import Grants
from .guardian import Guardian
from .hooks import Hooks
from .jobs import Jobs
from .log_streams import LogStreams
from .logs import Logs
from .resource_servers import ResourceServers
from .roles import Roles
from .rules import Rules
from .rules_configs import RulesConfigs
from .stats import Stats
from .tenants import Tenants
from .tickets import Tickets
from .user_blocks import UserBlocks
from .users import Users
from .users_by_email import UsersByEmail


class Auth0(object):
    """Provides easy access to all endpoint classes

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token
    """

    def __init__(self, domain, token):
        self.blacklists = Blacklists(domain, token)
        self.clients = Clients(domain, token)
        self.client_grants = ClientGrants(domain, token)
        self.custom_domains = CustomDomains(domain, token)
        self.connections = Connections(domain, token)
        self.device_credentials = DeviceCredentials(domain, token)
        self.emails = Emails(domain, token)
        self.email_templates = EmailTemplates(domain, token)
        self.grants = Grants(domain, token)
        self.guardian = Guardian(domain, token)
        self.hooks = Hooks(domain, token)
        self.jobs = Jobs(domain, token)
        self.logs = Logs(domain, token)
        self.log_streams = LogStreams(domain, token)
        self.resource_servers = ResourceServers(domain, token)
        self.roles = Roles(domain, token)
        self.rules = Rules(domain, token)
        self.rules_configs = RulesConfigs(domain, token)
        self.stats = Stats(domain, token)
        self.tenants = Tenants(domain, token)
        self.tickets = Tickets(domain, token)
        self.user_blocks = UserBlocks(domain, token)
        self.users = Users(domain, token)
        self.users_by_email = UsersByEmail(domain, token)



import unittest

from ...management.auth0 import Auth0
from ...management.blacklists import Blacklists
from ...management.client_grants import ClientGrants
from ...management.clients import Clients
from ...management.connections import Connections
from ...management.custom_domains import CustomDomains
from ...management.device_credentials import DeviceCredentials
from ...management.email_templates import EmailTemplates
from ...management.emails import Emails
from ...management.grants import Grants
from ...management.guardian import Guardian
from ...management.hooks import Hooks
from ...management.jobs import Jobs
from ...management.log_streams import LogStreams
from ...management.logs import Logs
from ...management.resource_servers import ResourceServers
from ...management.roles import Roles
from ...management.rules import Rules
from ...management.rules_configs import RulesConfigs
from ...management.stats import Stats
from ...management.tenants import Tenants
from ...management.tickets import Tickets
from ...management.user_blocks import UserBlocks
from ...management.users import Users
from ...management.users_by_email import UsersByEmail


class TestAuth0(unittest.TestCase):

    def setUp(self):
        self.domain = 'user.some.domain'
        self.token = 'a-token'
        self.a0 = Auth0(self.domain, self.token)

    def test_blacklists(self):
        self.assertIsInstance(self.a0.blacklists, Blacklists)

    def test_clients(self):
        self.assertIsInstance(self.a0.clients, Clients)

    def test_client_grants(self):
        self.assertIsInstance(self.a0.client_grants, ClientGrants)

    def test_custom_domains(self):
        self.assertIsInstance(self.a0.custom_domains, CustomDomains)

    def test_connections(self):
        self.assertIsInstance(self.a0.connections, Connections)

    def test_device_credentials(self):
        self.assertIsInstance(self.a0.device_credentials, DeviceCredentials)

    def test_emails(self):
        self.assertIsInstance(self.a0.emails, Emails)

    def test_email_templates(self):
        self.assertIsInstance(self.a0.email_templates, EmailTemplates)

    def test_grants(self):
        self.assertIsInstance(self.a0.grants, Grants)

    def test_guardian(self):
        self.assertIsInstance(self.a0.guardian, Guardian)

    def test_hooks(self):
        self.assertIsInstance(self.a0.hooks, Hooks)

    def test_jobs(self):
        self.assertIsInstance(self.a0.jobs, Jobs)

    def test_logs(self):
        self.assertIsInstance(self.a0.logs, Logs)

    def test_log_streams(self):
        self.assertIsInstance(self.a0.log_streams, LogStreams)

    def test_resource_servers(self):
        self.assertIsInstance(self.a0.resource_servers, ResourceServers)

    def test_roles(self):
        self.assertIsInstance(self.a0.roles, Roles)

    def test_rules(self):
        self.assertIsInstance(self.a0.rules, Rules)

    def test_rules_configs(self):
        self.assertIsInstance(self.a0.rules_configs, RulesConfigs)

    def test_stats(self):
        self.assertIsInstance(self.a0.stats, Stats)

    def test_tenants(self):
        self.assertIsInstance(self.a0.tenants, Tenants)

    def test_tickets(self):
        self.assertIsInstance(self.a0.tickets, Tickets)

    def test_user_blocks(self):
        self.assertIsInstance(self.a0.user_blocks, UserBlocks)

    def test_users(self):
        self.assertIsInstance(self.a0.users, Users)

    def test_users_by_email(self):
        self.assertIsInstance(self.a0.users_by_email, UsersByEmail)



from .rest import RestClient


class Logs(object):
    """Auth0 logs endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://{}/api/v2/logs'.format(self.domain)
        if id is not None:
            return '{}/{}'.format(url, id)
        return url

    def search(self, page=0, per_page=50, sort=None, q=None,
               include_totals=True, fields=None, from_param=None, take=None,
               include_fields=True):
        """Search log events.

        Args:
            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 50 results per page.

            sort (str, optional): The field to use for sorting.
                1 == ascending and -1 == descending. (e.g: date:1)
                When not set, the default value is up to the server.

            q (str, optional): Query in Lucene query string syntax.

            fields (list of str, optional): A list of fields to include or
                exclude from the result (depending on include_fields). Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise. Defaults to True.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to True.

            from_param (str, optional): Log Event Id to start retrieving logs. You can
                limit the amount of logs using the take parameter.

            take (int, optional): The total amount of entries to retrieve when
                using the from parameter. When not set, the default value is up to the server.

        See: https://auth0.com/docs/api/management/v2#!/Logs/get_logs
        """
        params = {
            'per_page': per_page,
            'page': page,
            'include_totals': str(include_totals).lower(),
            'sort': sort,
            'fields': fields and ','.join(fields) or None,
            'include_fields': str(include_fields).lower(),
            'q': q,
            'from': from_param,
            'take': take
        }
        return self.client.get(self._url(), params=params)

    def get(self, id):
        """Retrieves the data related to the log entry identified by id.

        Args:
            id (str): The log_id of the log to retrieve.

        See: https://auth0.com/docs/api/management/v2#!/Logs/get_logs_by_id
        """

        return self.client.get(self._url(id))



import unittest
import mock
from ...authentication.logout import Logout


class TestLogout(unittest.TestCase):

    @mock.patch('auth0.v3.authentication.logout.Logout.get')
    def test_logout(self, mock_get):

        g = Logout('my.domain.com')

        g.logout(client_id='cid',
                 return_to='rto')

        args, kwargs = mock_get.call_args
        self.assertEqual(args[0], 'https://my.domain.com/v2/logout?client_id=cid&returnTo=rto')

    @mock.patch('auth0.v3.authentication.logout.Logout.get')
    def test_federated_logout(self, mock_get):

        g = Logout('my.domain.com')

        g.logout(client_id='cid',
                 return_to='rto',
                 federated=True)

        args, kwargs = mock_get.call_args
        self.assertEqual(args[0], 'https://my.domain.com/v2/logout?federated&client_id=cid&returnTo=rto')



import warnings

from .rest import RestClient


class Users(object):
    """Auth0 users endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://{}/api/v2/users'.format(self.domain)
        if id is not None:
            return '{}/{}'.format(url, id)
        return url

    def list(self, page=0, per_page=25, sort=None, connection=None, q=None,
             search_engine=None, include_totals=True, fields=None,
             include_fields=True):
        """List or search users.

        Args:
            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 25 results per page.

            sort (str, optional): The field to use for sorting.
                1 == ascending and -1 == descending. (e.g: email:1)
                When not set, the default value is up to the server.

            connection (str, optional): Connection filter.

            q (str, optional): Query in Lucene query string syntax. Only fields
                in app_metadata, user_metadata or the normalized user profile
                are searchable.

            search_engine (str, optional): The version of the search_engine to use
                when querying for users. Will default to the latest version available.
                See: https://auth0.com/docs/users/search.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to True.

            fields (list of str, optional): A list of fields to include or
                exclude from the result (depending on include_fields). Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be include in the result, False otherwise. Defaults to True.

        See: https://auth0.com/docs/api/management/v2#!/Users/get_users
        """
        params = {
            'per_page': per_page,
            'page': page,
            'include_totals': str(include_totals).lower(),
            'sort': sort,
            'connection': connection,
            'fields': fields and ','.join(fields) or None,
            'include_fields': str(include_fields).lower(),
            'q': q,
            'search_engine': search_engine
        }
        return self.client.get(self._url(), params=params)

    def create(self, body):
        """Creates a new user.

        Args:
            body (dict): the attributes to set on the user to create.

        See: https://auth0.com/docs/api/v2#!/Users/post_users
        """
        return self.client.post(self._url(), data=body)

    def delete_all_users(self):
        """Deletes all users (USE WITH CAUTION).
        Deprecation: This endpoint is no longer available server-side.

        Args:
        """
        warnings.warn("DELETE all users endpoint is no longer available.", DeprecationWarning)
        return self.client.delete(self._url())

    def get(self, id, fields=None, include_fields=True):
        """Get a user.

        Args:
            id (str): The user_id of the user to retrieve.

            fields (list of str, optional): A list of fields to include or
                exclude from the result (depending on include_fields). Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise. Defaults to True.

        See: https://auth0.com/docs/api/management/v2#!/Users/get_users_by_id
        """
        params = {
            'fields': fields and ','.join(fields) or None,
            'include_fields': str(include_fields).lower()
        }

        return self.client.get(self._url(id), params=params)

    def delete(self, id):
        """Delete a user.

        Args:
            id (str): The user_id of the user to delete.

        See: https://auth0.com/docs/api/management/v2#!/Users/delete_users_by_id
        """
        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Update a user with the attributes passed in 'body'

        Args:
            id (str): The user_id of the user to update.

            body (dict): The attributes of the user to update.

        See: https://auth0.com/docs/api/v2#!/Users/patch_users_by_id
        """
        return self.client.patch(self._url(id), data=body)

    def list_roles(self, id, page=0, per_page=25, include_totals=True):
        """List the roles associated with a user.

        Args:
            id (str): The user's id.

            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 25 results per page.

            include_totals (bool, optional): True if the query summary is
               to be included in the result, False otherwise. Defaults to True.

        See https://auth0.com/docs/api/management/v2#!/Users/get_user_roles
        """
        params = {
            'per_page': per_page,
            'page': page,
            'include_totals': str(include_totals).lower()
        }

        url = self._url('{}/roles'.format(id))
        return self.client.get(url, params=params)

    def remove_roles(self, id, roles):
        """Removes an array of roles from a user.

        Args:
            id (str): The user's id.

            roles (list of str): A list of roles ids to unassociate from the user.

        See https://auth0.com/docs/api/management/v2#!/Users/delete_user_roles
        """
        url = self._url('{}/roles'.format(id))
        body = {'roles': roles}
        return self.client.delete(url, data=body)

    def add_roles(self, id, roles):
        """Associate an array of roles with a user.

        Args:
            id (str): The user's id.

            roles (list of str): A list of roles ids to associated with the user.

        See https://auth0.com/docs/api/management/v2#!/Users/post_user_roles
        """
        url = self._url('{}/roles'.format(id))
        body = {'roles': roles}
        return self.client.post(url, data=body)

    def list_permissions(self, id, page=0, per_page=25, include_totals=True):
        """List the permissions associated to the user.

        Args:
            id (str): The user's id.

            page (int, optional): The result's page number (zero based). By default,
               retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
               retrieves 25 results per page.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to True.

        See https://auth0.com/docs/api/management/v2#!/Users/get_permissions
        """

        params = {
            'per_page': per_page,
            'page': page,
            'include_totals': str(include_totals).lower()
        }
        url = self._url('{}/permissions'.format(id))
        return self.client.get(url, params=params)

    def remove_permissions(self, id, permissions):
        """Removes permissions from a user.

        Args:
            id (str): The user's id.

            permissions (list of str): A list of permission ids to unassociate from the user.

        See https://auth0.com/docs/api/management/v2#!/Users/delete_permissions
        """
        url = self._url('{}/permissions'.format(id))
        body = {'permissions': permissions}
        return self.client.delete(url, data=body)

    def add_permissions(self, id, permissions):
        """Assign permissions to a user.

        Args:
            id (str): The user's id.

            permissions (list of str): A list of permission ids to associated with the user.

        See https://auth0.com/docs/api/management/v2#!/Users/post_permissions
        """
        url = self._url('{}/permissions'.format(id))
        body = {'permissions': permissions}
        return self.client.post(url, data=body)

    def delete_multifactor(self, id, provider):
        """Delete a user's multifactor provider.

        Args:
            id (str): The user's id.

            provider (str): The multifactor provider. Supported values 'duo'
                or 'google-authenticator'.

        See: https://auth0.com/docs/api/management/v2#!/Users/delete_multifactor_by_provider
        """
        url = self._url('{}/multifactor/{}'.format(id, provider))
        return self.client.delete(url)

    def unlink_user_account(self, id, provider, user_id):
        """Unlink a user account

        Args:
            id (str): The user_id of the user identity.

            provider (str): The type of identity provider (e.g: facebook).

            user_id (str): The unique identifier for the user for the identity.

        See: https://auth0.com/docs/api/management/v2#!/Users/delete_user_identity_by_user_id
        """
        url = self._url('{}/identities/{}/{}'.format(id, provider, user_id))
        return self.client.delete(url)

    def link_user_account(self, user_id, body):
        """Link user accounts.

        Links the account specified in the body (secondary account) to the
        account specified by the id param of the URL (primary account).

        Args:
            id (str): The user_id of the primary identity where you are linking
                the secondary account to.

            body (dict): the attributes to send as part of this request.

        See: https://auth0.com/docs/api/v2#!/Users/post_identities
        """
        url = self._url('{}/identities'.format(user_id))
        return self.client.post(url, data=body)

    def regenerate_recovery_code(self, user_id):
        """Removes the current recovery token, generates and returns a new one

        Args:
            user_id (str):  The user_id of the user identity.

        See: https://auth0.com/docs/api/management/v2#!/Users/post_recovery_code_regeneration
        """
        url = self._url('{}/recovery-code-regeneration'.format(user_id))
        return self.client.post(url)

    def get_guardian_enrollments(self, user_id):
        """Retrieves all Guardian enrollments.

        Args:
            user_id (str):  The user_id of the user to retrieve.

        See: https://auth0.com/docs/api/management/v2#!/Users/get_enrollments
        """
        url = self._url('{}/enrollments'.format(user_id))
        return self.client.get(url)

    def get_log_events(self, user_id, page=0, per_page=50, sort=None,
                       include_totals=False):
        """Retrieve every log event for a specific user id.

        Args:
            user_id (str):  The user_id of the logs to retrieve.

            page (int, optional): The result's page number (zero based). By default,
                retrieves the first page of results.

            per_page (int, optional): The amount of entries per page. By default,
                retrieves 50 results per page.

            sort (str, optional):  The field to use for sorting. Use field:order
                where order is 1 for ascending and -1 for descending.
                For example date:-1
                When not set, the default value is up to the server.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to False.

        See: https://auth0.com/docs/api/management/v2#!/Users/get_logs_by_user
        """

        params = {
            'per_page': per_page,
            'page': page,
            'include_totals': str(include_totals).lower(),
            'sort': sort
        }

        url = self._url('{}/logs'.format(user_id))
        return self.client.get(url, params=params)

    def invalidate_remembered_browsers(self, user_id):
        """Invalidate all remembered browsers across all authentication factors for a user.

        Args:
            user_id (str):  The user_id to invalidate remembered browsers for.

        See: https://auth0.com/docs/api/management/v2#!/Users/post_invalidate_remember_browser
        """

        url = self._url('{}/multifactor/actions/invalidate-remember-browser'.format(user_id))
        return self.client.post(url)



from .rest import RestClient


class Jobs(object):
    """Auth0 jobs endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, path=None):
        url = 'https://{}/api/v2/jobs'.format(self.domain)
        if path is not None:
            return '{}/{}'.format(url, path)
        return url

    def get(self, id):
        """Retrieves a job. Useful to check its status.

        Args:
            id (str): The id of the job.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/get_jobs_by_id
        """
        return self.client.get(self._url(id))

    def get_failed_job(self, id):
        """Get failed job error details.

        Args:
            id (str): The id of the job.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/get_errors
        """
        url = self._url('{}/errors'.format(id))
        return self.client.get(url)

    def get_results(self, job_id):
        """Get results of a job

        Args:
            job_id (str): The id of the job.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/get_results
        """
        url = self._url('%s/results' % job_id)
        return self.client.get(url)

    def export_users(self, body):
        """Export all users to a file using a long running job.

        Check job status with get(). URL pointing to the export file will be
        included in the status once the job is complete.

        Args:
            body (dict): The details of the export users request.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/post_users_exports
        """
        return self.client.post(self._url('users-exports'), data=body)

    def import_users(self, connection_id, file_obj, upsert=False, send_completion_email=True, external_id=None):
        """Imports users to a connection from a file.

        Args:
            connection_id (str): The connection id of the connection to which
                users will be inserted.

            file_obj (file): A file-like object to upload. The format for
                this file is explained in: https://auth0.com/docs/bulk-import.

            upsert (bool, optional): When set to False, pre-existing users that match on email address, user ID, or username
                will fail. When set to True, pre-existing users that match on any of these fields will be updated, but
                only with upsertable attributes. Defaults to False.
                For a list of user profile fields that can be upserted during import, see the following article
                https://auth0.com/docs/users/references/user-profile-structure#user-profile-attributes.

            send_completion_email (bool, optional): When set to True, an email will be sent to notify the completion of this job.
                When set to False, no email will be sent. Defaults to True.

            external_id (str, optional):  Customer-defined ID.

        See: https://auth0.com/docs/api/management/v2#!/Jobs/post_users_imports
        """
        return self.client.file_post(self._url('users-imports'),
                                     data={'connection_id': connection_id,
                                           'upsert': str(upsert).lower(),
                                           'send_completion_email': str(send_completion_email).lower(),
                                           'external_id': external_id},
                                     files={'users': file_obj})

    def send_verification_email(self, body):
        """Send verification email.

        Send an email to the specified user that asks them to click a link to
        verify their email address.

        Args:
            body (dict): Details of verification email request.

        See: https://auth0.com/docs/api/v2#!/Jobs/post_verification_email
        """
        return self.client.post(self._url('verification-email'), data=body)



from .base import AuthenticationBase


class Social(AuthenticationBase):

    """Social provider's endpoints.

    Args:
        domain (str): Your auth0 domain (e.g: username.auth0.com)
    """

    def login(self, client_id, access_token, connection, scope='openid'):
        """Login using a social provider's access token

        Given the social provider's access_token and the connection specified,
        it will do the authentication on the provider and return a dict with
        the access_token and id_token. Currently, this endpoint only works for
        Facebook, Google, Twitter and Weibo.

        Args:
            client_id (str): application's client id.

            access_token (str): social provider's access_token.

            connection (str): connection type (e.g: 'facebook')

        Returns:
            A dict with 'access_token' and 'id_token' keys.
        """

        return self.post(
            'https://{}/oauth/access_token'.format(self.domain),
            data={
                'client_id': client_id,
                'access_token': access_token,
                'connection': connection,
                'scope': scope,
            }
        )



from .base import AuthenticationBase


class GetToken(AuthenticationBase):

    """/oauth/token related endpoints

    Args:
        domain (str): Your auth0 domain (e.g: username.auth0.com)
    """

    def authorization_code(self, client_id, client_secret, code,
                           redirect_uri, grant_type='authorization_code'):
        """Authorization code grant

        This is the OAuth 2.0 grant that regular web apps utilize in order
        to access an API. Use this endpoint to exchange an Authorization Code
        for a Token.

        Args:
            grant_type (str): Denotes the flow you're using. For authorization code
            use authorization_code

            client_id (str): your application's client Id

            client_secret (str): your application's client Secret

            code (str): The Authorization Code received from the /authorize Calls

            redirect_uri (str, optional): This is required only if it was set at
            the GET /authorize endpoint. The values must match

        Returns:
            access_token, id_token
        """

        return self.post(
            'https://{}/oauth/token'.format(self.domain),
            data={
                'client_id': client_id,
                'client_secret': client_secret,
                'code': code,
                'grant_type': grant_type,
                'redirect_uri': redirect_uri,
            }
        )

    def authorization_code_pkce(self, client_id, code_verifier, code,
                                redirect_uri, grant_type='authorization_code'):
        """Authorization code pkce grant

        This is the OAuth 2.0 grant that mobile apps utilize in order to access an API.
        Use this endpoint to exchange an Authorization Code for a Token.

        Args:
            grant_type (str): Denotes the flow you're using. For authorization code pkce
            use authorization_code

            client_id (str): your application's client Id

            code_verifier (str): Cryptographically random key that was used to generate
            the code_challenge passed to /authorize.

            code (str): The Authorization Code received from the /authorize Calls

            redirect_uri (str, optional): This is required only if it was set at
            the GET /authorize endpoint. The values must match

        Returns:
            access_token, id_token
        """

        return self.post(
            'https://{}/oauth/token'.format(self.domain),
            data={
                'client_id': client_id,
                'code_verifier': code_verifier,
                'code': code,
                'grant_type': grant_type,
                'redirect_uri': redirect_uri,
            }
        )

    def client_credentials(self, client_id, client_secret, audience,
                           grant_type='client_credentials'):
        """Client credentials grant

        This is the OAuth 2.0 grant that server processes utilize in
        order to access an API. Use this endpoint to directly request
        an access_token by using the Application Credentials (a Client Id and
        a Client Secret).

        Args:
            grant_type (str): Denotes the flow you're using. For client credentials
            use client_credentials

            client_id (str): your application's client Id

            client_secret (str): your application's client Secret

            audience (str): The unique identifier of the target API you want to access.

        Returns:
            access_token
        """

        return self.post(
            'https://{}/oauth/token'.format(self.domain),
            data={
                'client_id': client_id,
                'client_secret': client_secret,
                'audience': audience,
                'grant_type': grant_type,
            }
        )

    def login(self, client_id, client_secret, username, password, scope, realm,
              audience, grant_type='http://auth0.com/oauth/grant-type/password-realm'):
        """Calls /oauth/token endpoint with password-realm grant type


        This is the OAuth 2.0 grant that highly trusted apps utilize in order
        to access an API. In this flow the end-user is asked to fill in credentials
        (username/password) typically using an interactive form in the user-agent
        (browser). This information is later on sent to the client and Auth0.
        It is therefore imperative that the client is absolutely trusted with
        this information.

        Args:
            grant_type (str): Denotes the flow you're using. For password realm
            use http://auth0.com/oauth/grant-type/password-realm

            client_id (str): your application's client Id

            client_secret (str): your application's client Secret

            audience (str): The unique identifier of the target API you want to access.

            username (str): Resource owner's identifier

            password (str): resource owner's Secret

            scope(str): String value of the different scopes the client is asking for.
            Multiple scopes are separated with whitespace.

            realm (str): String value of the realm the user belongs.
            Set this if you want to add realm support at this grant.

        Returns:
            access_token, id_token
        """

        return self.post(
            'https://{}/oauth/token'.format(self.domain),
            data={
                'client_id': client_id,
                'username': username,
                'password': password,
                'realm': realm,
                'client_secret': client_secret,
                'scope': scope,
                'audience': audience,
                'grant_type': grant_type
            }
        )

    def refresh_token(self, client_id, client_secret, refresh_token, grant_type='refresh_token'):
        """Calls /oauth/token endpoint with refresh token grant type

        Use this endpoint to refresh an access token, using the refresh token you got during authorization.

        Args:
            grant_type (str): Denotes the flow you're using. For refresh token
            use refresh_token

            client_id (str): your application's client Id

            client_secret (str): your application's client Secret

            refresh_token (str): The refresh token returned from the initial token request.

        Returns:
            access_token, id_token
        """

        return self.post(
            'https://{}/oauth/token'.format(self.domain),
            data={
                'client_id': client_id,
                'client_secret': client_secret,
                'refresh_token': refresh_token,
                'grant_type': grant_type
            }
        )



from .rest import RestClient


class Grants(object):
    """Auth0 grants endpoints

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://%s/api/v2/grants' % self.domain
        if id is not None:
            return url + '/' + id
        return url

    def all(self, page=None, per_page=None, include_totals=False, extra_params=None):
        """Retrieves all grants.

        Args:
            page (int, optional): The result's page number (zero based). When not set,
               the default value is up to the server.

            per_page (int, optional): The amount of entries per page. When not set,
               the default value is up to the server.

            include_totals (bool, optional): True if the query summary is
               to be included in the result, False otherwise. Defaults to False.

           extra_params (dictionary, optional): The extra parameters to add to
               the request. The page, per_page, and include_totals values
               specified as parameters take precedence over the ones defined here.
            
        See: https://auth0.com/docs/api/management/v2#!/Grants/get_grants
        """
        params = extra_params or {}
        params.update({
            'page': page,
            'per_page': per_page,
            'include_totals': str(include_totals).lower()
        })

        return self.client.get(self._url(), params=params)

    def delete(self, id):
        """Deletes a grant.

        Args:
           id (str): The id of the grant to delete.

        See: https://auth0.com/docs/api/management/v2#!/Grants/delete_grants_by_id
        """
        url = self._url('%s' % (id))
        return self.client.delete(url)



from .rest import RestClient


class Rules(object):
    """Rules endpoint implementation.

    Args:
        domain (str): Your Auth0 domain, e.g: 'username.auth0.com'

        token (str): Management API v2 Token

        telemetry (bool, optional): Enable or disable Telemetry
            (defaults to True)

        timeout (float or tuple, optional): Change the requests
            connect and read timeout. Pass a tuple to specify
            both values separately or a float to set both to it.
            (defaults to 5.0 for both)
    """

    def __init__(self, domain, token, telemetry=True, timeout=5.0):
        self.domain = domain
        self.client = RestClient(jwt=token, telemetry=telemetry, timeout=timeout)

    def _url(self, id=None):
        url = 'https://{}/api/v2/rules'.format(self.domain)
        if id is not None:
            return '{}/{}'.format(url, id)
        return url

    def all(self, stage='login_success', enabled=True, fields=None,
            include_fields=True, page=None, per_page=None, include_totals=False):
        """Retrieves a list of all rules.

        Args:
            stage (str, optional):  Retrieves rules that match the execution stage.
                Defaults to login_success.

            enabled (bool, optional): If provided, retrieves rules that match
                the value, otherwise all rules are retrieved.

            fields (list, optional): A list of fields to include or exclude
                (depending on include_fields) from the result. Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise. Defaults to True.

            page (int, optional): The result's page number (zero based). When not set,
                the default value is up to the server.

            per_page (int, optional): The amount of entries per page. When not set,
                the default value is up to the server.

            include_totals (bool, optional): True if the query summary is
                to be included in the result, False otherwise. Defaults to False.

        See: https://auth0.com/docs/api/management/v2#!/Rules/get_rules
        """

        params = {
            'stage': stage,
            'fields': fields and ','.join(fields) or None,
            'include_fields': str(include_fields).lower(),
            'page': page,
            'per_page': per_page,
            'include_totals': str(include_totals).lower()
        }

        # since the default is True, this is here to disable the filter
        if enabled is not None:
            params['enabled'] = str(enabled).lower()

        return self.client.get(self._url(), params=params)

    def create(self, body):
        """Creates a new rule.

        Args:
            body (dict): Attributes for the newly created rule.

        See: https://auth0.com/docs/api/v2#!/Rules/post_rules
        """
        return self.client.post(self._url(), data=body)

    def get(self, id, fields=None, include_fields=True):
        """Retrieves a rule by its ID.

        Args:
            id (str): The id of the rule to retrieve.

            fields (list, optional): A list of fields to include or exclude
                (depending on include_fields) from the result. Leave empty to
                retrieve all fields.

            include_fields (bool, optional): True if the fields specified are
                to be included in the result, False otherwise. Defaults to True.

        See: https://auth0.com/docs/api/management/v2#!/Rules/get_rules_by_id
        """
        params = {'fields': fields and ','.join(fields) or None,
                  'include_fields': str(include_fields).lower()}
        return self.client.get(self._url(id), params=params)

    def delete(self, id):
        """Delete a rule.

        Args:
            id (str): The id of the rule to delete.

        See: https://auth0.com/docs/api/management/v2#!/Rules/delete_rules_by_id
        """
        return self.client.delete(self._url(id))

    def update(self, id, body):
        """Update an existing rule

        Args:
            id (str): The id of the rule to modify.

            body (dict): Attributes to modify.

        See: https://auth0.com/docs/api/v2#!/Rules/patch_rules_by_id
        """
        return self.client.patch(self._url(id), data=body)
